/*********************************************************************/
/*                                                                   */
/*  This Program Written By Paul Edwards.                            */
/*  Released to the public domain.                                   */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  unixio - Unix I/O functions written in terms of standard C       */
/*  functions.                                                       */
/*                                                                   */
/*********************************************************************/

/* Ok, so the main thing is that we are going beyond what C90
   specifies, so we need an extension - but minimal.
   
   We need to be able to open a directory and get the contents
   (just the names). opendir() will do that.
   
   To find out if a particular name is a directory, we can
   try doing a chdir() to get there.
   
   Also a getcwd() to preserve where we were so we can get back.
   
   Everything else is assumed to be a file and it can be tested
   by doing an fopen with rb.
   
   If you want to know the length, then use fseek.
   
   This minimizes the burden on the OS to provide more than C90.
   
   Don't expect to be able to do a rmdir() on a system like MVS 3.8J
   where directories can only be emulated as multiple PDSes. mkdir()
   would need preallocated PDSes for all the file types too. You
   need to trust the user to have set all that up and just report
   any issue.
   
   Somebody needs to maintain the state information when walking
   a tree. ftw() leaves it ambiguous who that is.
   
   Probably stat() can take a DIR entry and if appropriate to the
   system, determine whether it is a directory or file, and what
   the size is, to avoid having to do the chdir and fseek if
   required. Since this can potentially be expensive, the bits you
   want from stat() should be specified. ie use a different function,
   not stat(), and stat() will potentially be expensive, getting
   all information regardless of whether it will be used or not.
   
   Unclear whether standard should be set by DOS or Windows.
   Probably neither - needs a wrapper. DTA contains info for DOS.
   Currently no "stat" call implemented in PDOS/386. Probably
   not required. opendir and fopen should be sufficient.
   
   . and .. should not appear in any tree walk.
   
   fstat is not appropriate to use - remember what you opened yourself.
   
   All files/directories should be assumed read/write/executable,
   for a simple system.
   
   stat() could be implemented by calling opendir() to reduce
   OS burden.
*/

#include <stdio.h>
#include <time.h>
#include <string.h>

#include <windows.h>

#ifdef fileno
#undef fileno
#endif

#include "unixio.h"

static FILE *files[FOPEN_MAX];
/*char *environ[] = { NULL };*/

int open(const char *fnm, int mode, ...)
{
    int x;
    char modestr[20];
    int bin;
    int app;

    bin = mode & O_BINARY;
    if (bin)
    {
        mode &= ~O_BINARY;
    }
    app = mode & O_APPEND;
    if (app)
    {
        mode &= ~O_APPEND;
    }
    for (x = 3; x < FOPEN_MAX; x++)
    {
        if (files[x] == NULL)
        {
            break;
        }
    }
    if (x == FOPEN_MAX)
    {
        return (-1);
    }
    if (app)
    {
        strcpy(modestr, "a");
    }
    else if (mode == O_RDONLY)
    {
        strcpy(modestr, "r");
    }
    else if (mode == O_WRONLY)
    {
        strcpy(modestr, "w");
    }
    else if (mode == O_RDWR)
    {
        strcpy(modestr, "r+");
    }
    if (bin)
    {
        strcat(modestr, "b");
    }
    files[x] = fopen(fnm, modestr);
    if (files[x] == NULL)
    {
        return (-1);
    }
    return (x);
}

int read(int fno, void *buf, size_t bytes)
{
    size_t rb;

    if (fno < 3)
    {
        rb = fread(buf, 1, bytes, stdin);
    }
    else
    {
        rb = fread(buf, 1, bytes, files[fno]);
    }
    return ((int)rb);
}

int write(int fno, const void *buf, size_t bytes)
{
    size_t wb;

    if (fno == 1)
    {
        fwrite(buf, 1, bytes, stdout);
    }
    else if (fno == 2)
    {
        fwrite(buf, 1, bytes, stderr);
    }
    else if (fno > 2)
    {
        wb = fwrite(buf, 1, bytes, files[fno]);
    }
    return ((int)wb);
}

int close(int fno)
{
    if (fno >= 3)
    {
        fclose(files[fno]);
        files[fno] = NULL;
    }
    return (0);
}

long lseek(int fno, long int offset, int whence)
{
    return (fseek(files[fno], offset, whence));
}

FILE *fdopen(int fno, const char *mode)
{
    return (files[fno]);
}

char *getpwd()
{
    static char buf[FILENAME_MAX * 2];

    GetCurrentDirectory(sizeof buf, buf);
    return (buf);
}

char *getcwd(char *buf, int len)
{
    if (len != 0)
    {
        GetCurrentDirectory(len, buf);
    }
    return (buf);
}

void unlink(char *f)
{
    remove(f);
    return;
}


#if 0
/* return 0 if file/dir exists, -1 I guess otherwise */

int stat(char *f, struct stat *buf)
{
    FILE *fp;

    memset(buf, '\0', sizeof *buf);
    fp = fopen(f, "rb");
    if (fp == NULL)
    {
        errno = ENOENT;
        return (-1);
    }
    fseek(fp, 0, SEEK_END);
    buf->st_size = ftell(fp);
    buf->st_mode = 7; /* read, write, execute */
    fclose(fp);
    return (0);
}
#endif


int fileno(FILE *fp)
{
    return (0);
}

#if 0
int access(char *f, int n)
{
    FILE *fp;
    
    fp = fopen(f, "rb");
    if (fp == NULL) return (1);
    fclose(fp);
    return (0);
}
#endif

int fstat(int fh, struct stat *buf)
{
    memset(buf, '\0', sizeof *buf);
    buf->st_mode = 7; /* read, write, execute */
    return (0);
}

int pwait(int a, int *b, int c)
{
    return (0);
}

int putenv(const char *x)
{
    return (0);
}

char *mktemp(char *s)
{
    static long counter = 1;

    sprintf(s + strlen(s) - 6, "%06ld", counter);
    counter++;
    return (s);
}

int chdir(char *path)
{
    int ret;
    
    ret = SetCurrentDirectory(path);
    if (ret == 0) return (-1);
    return (0); /* success */
}

int rmdir(char *path)
{
    RemoveDirectory(path);
    return (0);
}

int mkdir(char *path, int permissions)
{
    CreateDirectory(path, 0);
    return (0);
}

FILE *popen(const char *s, const char *t)
{
    return (NULL);
}

int pclose(FILE *f)
{
    return (0);
}

int kill(int a, int b)
{
    return (0);
}

int mkstemp(char *buf)
{
    return (open(tmpnam(buf), O_WRONLY));
}

#if 0
DIR *opendir(const char *dirname)
{
    return (NULL);
}

struct dirent *readdir(DIR *dir)
{
    return (NULL);
}

int closedir(DIR *dir)
{
    return (0);
}
#endif

int execvp(const char *path, char **argv)
{
    return (0);
}

int chmod(char *x, int y)
{
    return (0);
}

int umask(int x)
{
    return (0);
}

int isatty(int x)
{
    return (0);
}

int pexecute()
{
    return (0);
}
