/*  msd_dir.c - portable directory routines
    Copyright (C) 1990 by Thorsten Ohl, td12@ddagsi3.bitnet

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.  */

/* Everything non trivial in this code is from: @(#)msd_dir.c 1.4
   87/11/06.  A public domain implementation of BSD directory routines
   for MS-DOS.  Written by Michael Rendell ({uunet,utai}michael@garfield),
   August 1897 */


#ifndef PUREISO
#include <io.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef PUREISO
#include <sys/types.h>
#include <sys/stat.h>
#else
#include <config.h>
#endif

#ifndef PUREISO
#include <dos.h>
#endif

#include <ndir.h>
extern void *xmalloc (size_t bytes);

static void free_dircontents (struct _dircontents *);

/* find ALL files! */
#define ATTRIBUTES	(_A_RDONLY | _A_HIDDEN | _A_SYSTEM | _A_SUBDIR)


#ifdef PUREISO

#include <windows.h>

#define _MAX_PATH (FILENAME_MAX * 2)

struct _finddata_t {
char name[FILENAME_MAX];
};

static HANDLE findHand;
static WIN32_FIND_DATA findData;

long _findfirst (char *name_buf, struct _finddata_t *find_buf)
{
    findHand = FindFirstFile(name_buf, &findData);
    if (findHand == INVALID_HANDLE_VALUE) return -1;

/* CVS does its own check */
#if 0
    if (strcmp(findData.cFileName, ".") == 0)
    {
        if (FindNextFile(findHand, &findData) == 0)
        {
            return -1;
        }
    }
    if (strcmp(findData.cFileName, "..") == 0)
    {
        if (FindNextFile(findHand, &findData) == 0)
        {
            return -1;
        }
    }
    if (strcmp(findData.cFileName, ".") == 0)
    {
        if (FindNextFile(findHand, &findData) == 0)
        {
            return -1;
        }
    }
#endif
    strcpy(find_buf->name, findData.cFileName);
    return 0;
}

int _findnext (long hFile, struct _finddata_t *find_buf)
{
    int ret;
    
    ret = FindNextFile(findHand, &findData);

/* CVS does its own check */
#if 0
    if ((ret != 0) && (strcmp(findData.cFileName, ".") == 0))
    {
        if (FindNextFile(findHand, &findData) == 0)
        {
            return -1;
        }
    }
    if ((ret != 0) && (strcmp(findData.cFileName, "..") == 0))
    {
        if (FindNextFile(findHand, &findData) == 0)
        {
            return -1;
        }
    }
    if ((ret != 0) && (strcmp(findData.cFileName, ".") == 0))
    {
        if (FindNextFile(findHand, &findData) == 0)
        {
            return -1;
        }
    }
#endif
    if (ret != 0)
    {
        strcpy(find_buf->name, findData.cFileName);
        return (0);
    }
    return (-1);
}

void _findclose(long hFile)
{
    FindClose(findHand);
    return;
}

#endif


DIR *
opendir (const char *name)
{
  struct _finddata_t find_buf;
  DIR *dirp;
  struct _dircontents *dp;
  char name_buf[_MAX_PATH + 1];
  char *slash = "";
  long hFile;

  if (!name)
    name = "";
  else if (*name)
    {
      const char *s;
      int l = strlen (name);

      s = name + l - 1;
      if ( !(l == 2 && *s == ':') && *s != '\\' && *s != '/')
	slash = "/";	/* save to insert slash between path and "*.*" */
    }

  strcat (strcat (strcpy (name_buf, name), slash), "*.*");

  dirp = (DIR *) xmalloc (sizeof (DIR));
  if (dirp == (DIR *)0)
    return (DIR *)0;

  dirp->dd_loc = 0;
  dirp->dd_contents = dirp->dd_cp = (struct _dircontents *) 0;

  if ((hFile = _findfirst (name_buf, &find_buf)) < 0)
    {
      free (dirp);
      return (DIR *)0;
    }

  do
    {
      dp = (struct _dircontents *) xmalloc (sizeof (struct _dircontents));
      if (dp == (struct _dircontents *)0)
	{
	  free_dircontents (dirp->dd_contents);
	  return (DIR *)0;
	}

      dp->_d_entry = xmalloc (strlen (find_buf.name) + 1);
      if (dp->_d_entry == (char *)0)
	{
	  free (dp);
	  free_dircontents (dirp->dd_contents);
	  return (DIR *)0;
	}

      if (dirp->dd_contents)
	dirp->dd_cp = dirp->dd_cp->_d_next = dp;
      else
	dirp->dd_contents = dirp->dd_cp = dp;

      strcpy (dp->_d_entry, find_buf.name);

      dp->_d_next = (struct _dircontents *)0;

    } while (!_findnext (hFile, &find_buf));

  dirp->dd_cp = dirp->dd_contents;

  _findclose(hFile);

  return dirp;
}


void
closedir (DIR *dirp)
{
  free_dircontents (dirp->dd_contents);
  free ((char *) dirp);
}


struct direct *
readdir (DIR *dirp)
{
  static struct direct dp;

  if (dirp->dd_cp == (struct _dircontents *)0)
    return (struct direct *)0;
  dp.d_namlen = dp.d_reclen =
    strlen (strcpy (dp.d_name, dirp->dd_cp->_d_entry));
#if 0 /* JB */
  strlwr (dp.d_name);		/* JF */
#endif
  dp.d_ino = 0;
  dirp->dd_cp = dirp->dd_cp->_d_next;
  dirp->dd_loc++;

  return &dp;
}


void
seekdir (DIR *dirp, long off)
{
  long i = off;
  struct _dircontents *dp;

  if (off < 0)
    return;
  for (dp = dirp->dd_contents; --i >= 0 && dp; dp = dp->_d_next)
    ;
  dirp->dd_loc = off - (i + 1);
  dirp->dd_cp = dp;
}


long
telldir (DIR *dirp)
{
  return dirp->dd_loc;
}


/* Garbage collection */

static void
free_dircontents (struct _dircontents *dp)
{
  struct _dircontents *odp;

  while (dp)
    {
      if (dp->_d_entry)
	free (dp->_d_entry);
      dp = (odp = dp)->_d_next;
      free (odp);
    }
}

#ifdef PUREISO
int stat(char *f, struct stat *buf)
{
    DIR *d;
    FILE *fp;
    
    memset(buf, '\0', sizeof *buf);
    d = opendir(f);
    if ((d != NULL) || (GetLastError() == ERROR_FILE_NOT_FOUND))
    {
        if (d != NULL)
        {
            closedir(d);
        }
        buf->st_mode = 7; /* read, write, execute */
        buf->st_mode |= S_IFDIR;
        return 0;
    }
    fp = fopen(f, "rb");
    if (fp == NULL)
    {
        return -1;
    }
    fclose(fp);
    buf->st_mode = 7; /* read, write, execute */
    return 0;
}

int access(char *f, int n)
{
    DIR *d;
    FILE *fp;
    
    d = opendir(f);
    if (d != NULL)
    {
        closedir(d);
        return 0;
    }
    fp = fopen(f, "rb");
    if (fp == NULL)
    {
        return -1;
    }
    fclose(fp);
    return 0;

}

#endif


#ifdef TEST

void main (int argc, char *argv[]);

void
main (int argc, char *argv[])
{
  static DIR *directory;
  struct direct *entry = (struct direct *)0;

  char *name = "";

  if (argc > 1)
    name = argv[1];

  directory = opendir (name);

  if (!directory)
    {
      fprintf (stderr, "can't open directory `%s'.\n", name);
      exit (2);
    }

  while (entry = readdir (directory))
    printf ("> %s\n", entry->d_name);

  printf ("done.\n");
}

#endif /* TEST */

/* 
 * Local Variables:
 * mode:C
 * ChangeLog:ChangeLog
 * compile-command:make
 * End:
 */
